@@ -27,13 +27,13 @@
 
 ;; `easy-kill' aims to be a drop-in replacement for `kill-ring-save'.
 ;;
-;; To use: (global-set-key [remap kill-ring-save] 'easy-kill)
+;; To use: (global-set-key [remap kill-ring-save] #'easy-kill)
 
 ;; `easy-mark' is similar to `easy-kill' but marks the region
 ;; immediately. It can be a handy replacement for `mark-sexp' allowing
 ;; `+'/`-' to do list-wise expanding/shrinking.
 ;;
-;; To use: (global-set-key [remap mark-sexp] 'easy-mark)
+;; To use: (global-set-key [remap mark-sexp] #'easy-mark)
 
 ;; Please send bug reports or feature requests to:
 ;;      https://github.com/leoliu/easy-kill/issues
@@ -54,10 +54,7 @@
   "A list of (CHAR THING APPEND).
 CHAR is used immediately following `easy-kill' to select THING.
 APPEND is optional and if non-nil specifies the separator (a
-string) for appending current selection to previous kill.
-
-Note: each element can also be (CHAR . THING) but this is
-deprecated."
+string) for appending current selection to previous kill."
   :type '(repeat (list character symbol
                        (choice string (const :tag "None" nil))))
   :group 'killing)
@@ -72,16 +69,39 @@ deprecated."
   :type '(choice (const :tag "None" nil) key-sequence)
   :group 'killing)
 
-(defcustom easy-kill-try-things '(url email line)
+(defcustom easy-kill-try-things '(url email uuid line)
   "A list of things for `easy-kill' to try."
   :type '(repeat symbol)
   :group 'killing)
 
-(defcustom easy-mark-try-things '(url email sexp)
+(defcustom easy-kill-exit-hook nil
+  "A list of functions to be called when `easy-kill' exits.
+
+At the moment they are run, the current candidate is still accessible
+and `this-command' has the command that is to put an end to the
+easy-kill mode."
+  :type '(repeat function)
+  :group 'killing)
+
+(defcustom easy-mark-try-things '(url email uuid sexp)
   "A list of things for `easy-mark' to try."
   :type '(repeat symbol)
   :group 'killing)
 
+(defcustom easy-mark-exit-hook nil
+  "A list of functions to be called when `easy-mark' exits.
+
+At the moment they are run, the current candidate is still accessible
+and `this-command' has the command that is to put an end to the
+easy-mark mode."
+  :type '(repeat function)
+  :group 'killing)
+
+(defcustom easy-dup-try-things '(line)
+  "A list of things for `easy-dup-before' and `easy-dup-after' to try."
+  :type '(repeat symbol)
+  :group 'killing)
+
 (defface easy-kill-selection '((t (:inherit secondary-selection)))
   "Faced used to highlight kill candidate."
   :group 'killing)
@@ -92,21 +112,21 @@ deprecated."
 
 (defvar easy-kill-base-map
   (let ((map (make-sparse-keymap)))
-    (define-key map "-" 'easy-kill-shrink)
-    (define-key map "+" 'easy-kill-expand)
-    (define-key map "=" 'easy-kill-expand)
-    (define-key map " " 'easy-kill-cycle)
-    (define-key map "@" 'easy-kill-append)
+    (define-key map "-" #'easy-kill-shrink)
+    (define-key map "+" #'easy-kill-expand)
+    (define-key map "=" #'easy-kill-expand)
+    (define-key map " " #'easy-kill-cycle)
+    (define-key map "@" #'easy-kill-append)
     ;; Note: didn't pick C-h because it is a very useful prefix key.
-    (define-key map "?" 'easy-kill-help)
-    (define-key map [remap set-mark-command] 'easy-kill-mark-region)
-    (define-key map [remap kill-region] 'easy-kill-region)
-    (define-key map [remap delete-region] 'easy-kill-delete-region)
-    (define-key map [remap keyboard-quit] 'easy-kill-abort)
+    (define-key map "?" #'easy-kill-help)
+    (define-key map [remap set-mark-command] #'easy-kill-mark-region)
+    (define-key map [remap kill-region] #'easy-kill-region)
+    (define-key map [remap delete-region] #'easy-kill-delete-region)
+    (define-key map [remap keyboard-quit] #'easy-kill-abort)
     (define-key map [remap exchange-point-and-mark]
-      'easy-kill-exchange-point-and-mark)
+      #'easy-kill-exchange-point-and-mark)
     (mapc (lambda (d)
-            (define-key map (number-to-string d) 'easy-kill-digit-argument))
+            (define-key map (number-to-string d) #'easy-kill-digit-argument))
           (number-sequence 0 9))
     map))
 
@@ -124,7 +144,7 @@ deprecated."
 Do nothing if `easy-kill-inhibit-message' is non-nil."
   (unless easy-kill-inhibit-message
     (let (message-log-max)
-      (apply 'message format-string args))))
+      (apply #'message format-string args))))
 
 (defun easy-kill-trim (s &optional how)
   (let ((wchars "[ \t\n\r\f\v]*"))
@@ -166,11 +186,10 @@ The value is the function's symbol if non-nil."
     (set-keymap-parent map easy-kill-base-map)
     (when easy-kill-unhighlight-key
       (with-demoted-errors "easy-kill-unhighlight-key: %S"
-        (define-key map easy-kill-unhighlight-key 'easy-kill-unhighlight)))
-    (mapc (lambda (c)
-            ;; (define-key map (vector meta-prefix-char c) 'easy-kill-select)
-            (define-key map (char-to-string c) 'easy-kill-thing))
-          (mapcar 'car easy-kill-alist))
+        (define-key map easy-kill-unhighlight-key #'easy-kill-unhighlight)))
+    (dolist (c easy-kill-alist)
+      ;; (define-key map (vector meta-prefix-char (car c)) #'easy-kill-select)
+      (define-key map (char-to-string (car c)) #'easy-kill-thing))
     map))
 
 (defun easy-kill--fmt (x y &optional z)
@@ -199,11 +218,8 @@ The value is the function's symbol if non-nil."
     (princ "\n")
     (princ (easy-kill--fmt "---" "-----" "---------"))
     (princ "\n\n")
-    (princ (mapconcat (lambda (x) (pcase x
-                                    (`(,c ,thing ,sep)
-                                     (easy-kill--fmt c thing sep))
-                                    ((or `(,c ,thing) `(,c . ,thing))
-                                     (easy-kill--fmt c thing))))
+    (princ (mapconcat (pcase-lambda (`(,c ,thing ,sep))
+                        (easy-kill--fmt c thing sep))
                       easy-kill-alist "\n"))
     (princ "\n\n")
     (princ (substitute-command-keys "\\{easy-kill-base-map}"))))
@@ -278,7 +294,7 @@ non-zero length, it is the string covered by the overlay.
 Otherwise, it is the value of the overlay's candidate property."
   (with-current-buffer (easy-kill-get buffer)
     (or (pcase (easy-kill-get bounds)
-          (`(,_x . ,_x) (easy-kill-get candidate))
+          (`(,x . ,x) (ignore x) (easy-kill-get candidate))
           (`(,beg . ,end) (filter-buffer-substring beg end)))
         "")))
 
@@ -369,16 +385,14 @@ A thing is opted out of cycling if in `easy-kill-cycle-ignored'."
       (easy-kill-cycle next))))
 
 (defun easy-kill-cycle-next (thing depth)
-  (cl-flet ((thing-name (thing)
-              (if (symbolp (cdr thing)) (cdr thing) (cl-second thing))))
-    (let ((next (thing-name
-                 (car (or (cl-loop for (head . tail) on easy-kill-alist
-                                   when (eq thing (thing-name head))
-                                   return tail)
-                          easy-kill-alist)))))
-      (cond ((not (memq next easy-kill-cycle-ignored)) next)
-            ((> depth 0) (easy-kill-cycle-next next (1- depth)))
-            (t (user-error "Nothing to cycle"))))))
+  (let ((next (cl-second
+               (car (or (cl-loop for (head . tail) on easy-kill-alist
+                                 when (eq thing (cl-second head))
+                                 return tail)
+                        easy-kill-alist)))))
+    (cond ((not (memq next easy-kill-cycle-ignored)) next)
+          ((> depth 0) (easy-kill-cycle-next next (1- depth)))
+          (t (user-error "Nothing to cycle")))))
 
 (defun easy-kill-digit-argument (n)
   "Expand selection by N number of things.
@@ -452,7 +466,7 @@ checked."
                         (`1 end))))
            (new-front (save-excursion
                         (goto-char front)
-                        (with-demoted-errors
+                        (with-demoted-errors "%S"
                           (dotimes (_ (abs n))
                             (easy-kill-thing-forward-1 thing step)))
                         (point))))
@@ -465,9 +479,7 @@ checked."
 (defun easy-kill-thing (&optional thing n inhibit-handler)
   ;; N can be -, + and digits
   (interactive
-   (list (pcase (assq last-command-event easy-kill-alist)
-           (`(,_ ,th . ,_) th)
-           (`(,_ . ,th) th))
+   (list (cl-second (assq last-command-event easy-kill-alist))
          (prefix-numeric-value current-prefix-arg)))
   (let* ((thing (or thing (easy-kill-get thing)))
          (n (or n 1))
@@ -495,26 +507,37 @@ checked."
       (easy-kill-adjust-candidate (easy-kill-get thing)))))
 
 (easy-kill-defun easy-kill-abort ()
+  "Abort `easy-kill'/`easy-mark'."
   (interactive)
-  (when (easy-kill-get mark)
-    ;; The after-string may interfere with `goto-char'.
-    (overlay-put (easy-kill-get origin-indicator) 'after-string nil)
-    (goto-char (easy-kill-get origin))
-    (setq deactivate-mark t))
+  (and (easy-kill-get mark)
+       (easy-kill-terminate))
   (ding))
 
+(easy-kill-defun easy-kill-terminate ()
+  "Terminate `easy-mark', or save selection and terminate `easy-kill'."
+  (interactive)
+  (if (easy-kill-get mark)
+      (progn
+        ;; The after-string may interfere with `goto-char'.
+        (overlay-put (easy-kill-get origin-indicator) 'after-string nil)
+        (goto-char (easy-kill-get origin))
+        (setq deactivate-mark t))
+    (pcase (easy-kill-get bounds)
+      (`(,x . ,x) (ignore x))
+      (`(,beg . ,end) (kill-ring-save beg end)))))
+
 (easy-kill-defun easy-kill-region ()
   "Kill current selection and exit."
   (interactive "*")
   (pcase (easy-kill-get bounds)
-    (`(,_x . ,_x) (easy-kill-echo "Empty region"))
+    (`(,x . ,x) (ignore x) (easy-kill-echo "Empty region"))
     (`(,beg . ,end) (kill-region beg end))))
 
 (easy-kill-defun easy-kill-mark-region ()
+  "Mark current selection and exit."
   (interactive)
   (pcase (easy-kill-get bounds)
-    (`(,_x . ,_x)
-     (easy-kill-echo "Empty region"))
+    (`(,x . ,x) (ignore x) (easy-kill-echo "Empty region"))
     (`(,beg . ,end)
      (pcase (if (eq (easy-kill-get mark) 'end)
                 (list end beg) (list beg end))
@@ -567,8 +590,11 @@ checked."
                           (command-remapping cmd nil (list map)))))
                (ignore
                 (easy-kill-destroy-candidate)
-                (unless (or (easy-kill-get mark) (easy-kill-exit-p this-command))
-                  (easy-kill-save-candidate))))
+                (if (easy-kill-get mark)
+                    (run-hooks 'easy-mark-exit-hook)
+                  (or (easy-kill-exit-p this-command)
+                      (easy-kill-save-candidate))
+                  (run-hooks 'easy-kill-exit-hook))))
          (error (message "%s:%s" this-command (error-message-string err))
                 nil))))))
 
@@ -601,7 +627,7 @@ Temporally activate additional key bindings as follows:
     (easy-kill-activate-keymap)))
 
 ;;;###autoload
-(defalias 'easy-mark-sexp 'easy-mark
+(defalias 'easy-mark-sexp #'easy-mark
   "Use `easy-mark' instead. The alias may be removed in future.")
 
 ;;;###autoload
@@ -615,6 +641,63 @@ Temporally activate additional key bindings as follows:
       (setf (easy-kill-get thing) 'sexp)
       (easy-kill-thing 'sexp n))))
 
+(declare-function rectangle--duplicate-right "rect" (n displacement))
+
+;;;###autoload
+(defun easy-dup (&optional n before)
+  "Insert a copy of the current selection after it, or before it if BEFORE.
+When not in easy-kill/easy-mark, use the active region if available, or
+enter easy-mark using `easy-dup-try-things' to select something to
+duplicate.  `rectangle-mark-mode' is also supported.  N specifies the
+number of copies to insert."
+  (interactive "*p")
+  (or
+   (pcase (if easy-kill-candidate (easy-kill-get bounds) '(nil . nil))
+     (`(,x . ,x)
+      (ignore x)
+      (cond
+       ((bound-and-true-p rectangle-mark-mode)
+        (rectangle--duplicate-right n (if before n 0))
+        t)
+       ((use-region-p)
+        (let* ((beg (region-beginning))
+               (end (region-end))
+               (text (buffer-substring beg end)))
+          (save-excursion
+            (goto-char (if before beg end))
+            (if before
+                (dotimes (_ (or n 1)) (insert-before-markers text))
+              (dotimes (_ (or n 1)) (insert text)))))
+        t)
+       (t (let ((easy-mark-try-things easy-dup-try-things))
+            (easy-mark 1)
+            nil)))))
+   (pcase (easy-kill-get bounds)
+     (`(,x . ,x) (ignore x) (easy-kill-echo "Empty region"))
+     (`(,beg . ,end)
+      (let ((text (buffer-substring beg end)))
+        (save-excursion
+          (goto-char (if before beg end))
+          (if before
+              (dotimes (_ (or n 1)) (insert-before-markers text))
+            (dotimes (_ (or n 1)) (insert text))))
+        (and before
+             (setf (easy-kill-get origin) (easy-kill-get start)))))))
+  (setq deactivate-mark nil))
+
+;;;###autoload
+(defalias 'easy-dup-after #'easy-dup)
+
+;;;###autoload
+(defun easy-dup-before (&optional n)
+  "Insert a copy of the current selection before it.
+When not in easy-kill/easy-mark, use the active region if available, or
+enter easy-mark using `easy-dup-try-things' to select something to
+duplicate.  `rectangle-mark-mode' is also supported.  N specifies the
+number of copies to insert."
+  (interactive "*p")
+  (easy-dup n t))
+
 ;;;; Extended things
 
 ;;; Handler for `buffer-file-name'.
@@ -797,8 +880,8 @@ inspected."
     (n (ignore-errors
          (dotimes (_ (abs n))
            (pcase (list (point) (easy-kill-bounds-of-thing-at-point 'list))
-             (`(,_beg (,_beg . ,_)) (org-up-element))
-             (`(,_ (,beg . ,_))     (goto-char beg)))))
+             (`(,beg (,beg . ,_)) (ignore beg) (org-up-element))
+             (`(,_ (,beg . ,_))   (goto-char beg)))))
        (when (cl-plusp n)
          (goto-char (cdr (easy-kill-bounds-of-thing-at-point 'list)))))))
 
